#include <stdio.h>
#include <math.h>
#include "pico/stdlib.h"
#include "pico/binary_info.h"
#include "pico/multicore.h"         // für critical_section für DI / EI
#include "hardware/gpio.h"
#include "hardware/dma.h"
#include "hardware/pio.h"
#include "hardware/pio_instructions.h"
#include "hardware/structs/pwm.h"   // for example waveforms to analyse
#include "XY2-100.pio.h"
#include "cdefs.h"

#include "LogicAnalyzer.h"





void logic_analyser_init(PIO pio, uint sm, uint pin_base, uint16_t pin_count, float div)
{
    // Load a program to capture n pins. This is just a single `in pins, n`
    // instruction with a wrap.
    uint16_t capture_prog_instr[] = { uint16_t(pio_encode_in(pio_pins, pin_count)) };
    struct pio_program capture_prog = {
            .instructions = capture_prog_instr,
            .length = 1,
            .origin = -1
    };
    uint offset = pio_add_program(pio, &capture_prog);

    // Configure state machine to loop over this `in` instruction forever,
    // with autopush enabled.
    pio_sm_config c = pio_get_default_sm_config();
    sm_config_set_in_pins(&c, pin_base);
    sm_config_set_wrap(&c, offset, offset);
    sm_config_set_clkdiv(&c, div);
    sm_config_set_in_shift(&c, true, true, 32);
    sm_config_set_fifo_join(&c, PIO_FIFO_JOIN_RX);
    pio_sm_init(pio, sm, offset, &c);
}

void logic_analyser_arm(PIO pio, uint sm, uint dma_chan, uint32_t *capture_buf, size_t capture_size_words,
                        uint trigger_pin, bool trigger_level)
{
    pio_sm_set_enabled(pio, sm, false);
    pio_sm_clear_fifos(pio, sm);

    dma_channel_config c = dma_channel_get_default_config(dma_chan);
    channel_config_set_read_increment(&c, false);
    channel_config_set_write_increment(&c, true);
    channel_config_set_dreq(&c, pio_get_dreq(pio, sm, false));

    dma_channel_configure(dma_chan, &c,
        capture_buf,        // Destinatinon pointer
        &pio->rxf[sm],      // Source pointer
        capture_size_words, // Number of transfers
        true                // Start immediately
    );

    pio_sm_exec(pio, sm, pio_encode_wait_gpio(trigger_level, trigger_pin));
    pio_sm_set_enabled(pio, sm, true);
}

void print_capture_buf(const uint32_t *buf, uint pin_base, uint pin_count, uint32_t n_samples)
{
    while(n_samples > 128)
    {
        print_capture_buf(buf, pin_base, pin_count, 128);
        buf += pin_count * 128 / 32;
        n_samples -= 128;
    }

    // Display the capture buffer in text form, like this:
    // 00: __--__--__--__--__--__--
    // 01: ____----____----____----
    printf("Capture:\n");
    for (int pin = 0; pin < pin_count; ++pin)
    {
        printf("%02d: ", pin + pin_base);
        for (int sample = 0; sample < n_samples; ++sample)
        {
            uint bit_index = pin + sample * pin_count;
            bool level = (buf[bit_index / 32] >> (bit_index % 32)) & 1u;
            printf(level ? "-" : "_");
        }
        printf("\n");
    }
}


