// copyright 2021 kio@little-bat.de
// BSD 2-clause license


#include <stdio.h>
#include <math.h>
#include "pico/stdlib.h"
#include "pico/binary_info.h"
#include "hardware/gpio.h"
#include "hardware/dma.h"
#include "hardware/pio.h"
#include "hardware/pio_instructions.h"
#include "hardware/structs/pwm.h"   // for example waveforms to analyse
//#include "pico/multicore.h"         // für critical_section für DI / EI
#include "cdefs.h"
#include "XY2.h"
#include "XY2-100.pio.h"


//static inline FLOAT sin (FLOAT a) { return sinf(a); }
//static inline FLOAT cos (FLOAT a) { return cosf(a); }
static const FLOAT pi = 3.1415926538;


#define pio PIO_XY

//static
uint XY2::heart_beat_counter = 1000;
uint XY2::heart_beat_state   = 0;
int XY2::x0 = 0;
int XY2::y0 = 0;


XY2::XY2()
{
    init_clock();
    init_data();
    init_laser();
    send_data(0,0,off);	// x0=y0=0;
}

//void XY2::init_clock()
//{
//    // setup clock+sync program on state machine 'sm_clock' in pio block 'pio'
//    // signals are sent to pins 'pin_clock' .. 'pin_clock+3'
//
//    // set initial pin states:
//    pio_sm_set_pins_with_mask(pio, sm_clock, 0b1111, 0x0fu << pin_clock);
//
//    // set i/o direction for pins used by sm:
//    pio_sm_set_consecutive_pindirs(pio, sm_clock, pin_clock, 4, true/*output*/);
//
//    // set gpio to use pio output for pins: (gpio_set_function)
//    for(uint i=0;i<4;i++) pio_gpio_init(pio, pin_clock+i);
//
//    uint offset = pio_add_program(pio, &XY2_100_clock_program);
//    program_base_clock = offset;
//    pio_sm_config c = XY2_100_clock_program_get_default_config(offset);
//
//    sm_config_set_set_pins(&c, pin_clock, 4);    // pins to use in SET instr
//    sm_config_set_clkdiv(&c, float(MAIN_CLOCK) / XY2_100_clock_sm_clock);
//    pio_sm_init(pio, sm_clock, offset + XY2_100_clock_offset_start, &c);
//}

void XY2::init_clock()
{
    // setup clock+sync program on state machine 'sm_clock' in pio block 'pio'
    // signals are sent to pins 'pin_clock' .. 'pin_clock+3'

    // set initial pin states in the pio:
    pio_sm_set_pins_with_mask(pio, sm_clock, 0b1111 << pin_clock, 0x0fu << pin_clock);

    // set i/o direction for pins in the pio:
    pio_sm_set_consecutive_pindirs(pio, sm_clock, pin_clock, 4, true/*output*/);

    // set gpio to use pio output for pins:
    for(uint i=0;i<4;i++) pio_gpio_init(pio, pin_clock+i);

    // load program into pio:
    uint offset = pio_add_program(pio, &xy2_clock_program);
    program_base_clock = offset;

    // configure the SM:
    pio_sm_config c = xy2_clock_program_get_default_config(offset);
    sm_config_set_sideset_pins(&c, pin_clock);    // pins to use in SET instr
    sm_config_set_clkdiv(&c, float(MAIN_CLOCK) / XY2_SM_CLOCK);
    pio_sm_init(pio, sm_clock, offset + xy2_clock_offset_start, &c);
}

void XY2::init_data ()
{
    // setup x and y data programs on state machine 'sm' in pio block 'pio'
    // data signals are sent to pins 'pin_base' .. 'pin_base+3'
    // pin 'sync_pin' is used to synchronize the data state machines for 'pull vs. repeat' decision

    // set initial pin states in pio:
    pio_sm_set_pins_with_mask(pio, sm_x, 0b11u << pin_x, 0x03u << pin_x);
    pio_sm_set_pins_with_mask(pio, sm_y, 0b11u << pin_y, 0x03u << pin_y);
    //pio_sm_set_pins_with_mask(pio, sm_y, 0b1u << pin_sync_xy, 0b1u << pin_sync_xy);

    // set i/o direction for pins in pio:
    pio_sm_set_consecutive_pindirs(pio, sm_x, pin_x, 2, true/*output*/);
    pio_sm_set_consecutive_pindirs(pio, sm_y, pin_y, 2, true/*output*/);
    //pio_sm_set_consecutive_pindirs(pio, sm_y, pin_sync_xy, 1, true/*output*/);

    // set gpio to use pio output for pins: (gpio_set_function)
    pio_gpio_init(pio, pin_x);
    pio_gpio_init(pio, pin_x+1);
    pio_gpio_init(pio, pin_y);
    pio_gpio_init(pio, pin_y+1);
    //pio_gpio_init(pio, pin_sync_xy);

    // load program into pio:
    uint offset = pio_add_program(pio, &xy2_100_data_program);
    program_base_data = offset;

    // configure state machines:
    pio_sm_config c = xy2_100_data_program_get_default_config(offset);
    sm_config_set_clkdiv(&c, float(MAIN_CLOCK) / XY2_SM_CLOCK);
    sm_config_set_out_shift(&c, false/*shift left not right*/, false /*!autopull*/, 16+16 /*pull_threshold*/);
	sm_config_set_fifo_join(&c, PIO_FIFO_JOIN_TX);
    //sm_config_set_mov_status(&c, STATUS_TX_LESSTHAN, 1);
    //sm_config_set_out_pins(&c, pin_sync_xy, 1);        // pins to use in OUT instr and MOV dest.
    //sm_config_set_in_pins (&c, pin_sync_xy);           // pins to use in IN instr  and MOV source
    sm_config_set_jmp_pin(&c, pin_sync_xy);

    sm_config_set_sideset_pins(&c, pin_x);      // pins for 'side setting' of x+ and x-
    pio_sm_init(pio, sm_x, offset + xy2_100_data_offset_start, &c);

    sm_config_set_sideset_pins(&c, pin_y);      // pins for 'side setting' of y+ and y-
    pio_sm_init(pio, sm_y, offset + xy2_100_data_offset_start, &c);
}

//void XY2::init_data ()
//{
//    // setup x and y data programs on state machine 'sm' in pio block 'pio'
//    // data signals are sent to pins 'pin_base' .. 'pin_base+3'
//    // pin 'sync_pin' is used to synchronize the data state machines for 'pull vs. repeat' decision
//
//    // set initial pin states:
//    pio_sm_set_pins_with_mask(pio, sm_x, 0b11u, 0x03u << pin_x);
//    pio_sm_set_pins_with_mask(pio, sm_y, 0b11u, 0x03u << pin_y);
//    //pio_sm_set_pins_with_mask(pio, sm_x, 0b1u, 0b1u << pin_sync_xy);
//    pio_sm_set_pins_with_mask(pio, sm_y, 0b1u, 0b1u << pin_sync_xy);
//
//    // set i/o direction for pins used by sm:
//    pio_sm_set_consecutive_pindirs(pio, sm_x, pin_x, 2, true/*output*/);
//    pio_sm_set_consecutive_pindirs(pio, sm_y, pin_y, 2, true/*output*/);
//    //pio_sm_set_consecutive_pindirs(pio, sm_x, pin_sync_xy, 1, true/*output*/);
//    pio_sm_set_consecutive_pindirs(pio, sm_y, pin_sync_xy, 1, true/*output*/);
//
//    // set gpio to use pio output for pins: (gpio_set_function)
//    pio_gpio_init(pio, pin_x);
//    pio_gpio_init(pio, pin_x+1);
//    pio_gpio_init(pio, pin_y);
//    pio_gpio_init(pio, pin_y+1);
//    //pio_gpio_init(pio, pin_sync_xy);
//
//    // load program into pio:
//    uint offset = pio_add_program(pio, &xy2_100_data_program);
//    program_base_data = offset;
//
//    // configure state machines:
//    pio_sm_config c = xy2_100_data_program_get_default_config(offset);
//    sm_config_set_clkdiv(&c, float(MAIN_CLOCK) / XY2_SM_CLOCK);
//    sm_config_set_out_shift(&c, false/*shift left not right*/, false /*!autopull*/, 16+16 /*pull_threshold*/);
//	sm_config_set_fifo_join(&c, PIO_FIFO_JOIN_TX);
//    //sm_config_set_mov_status(&c, STATUS_TX_LESSTHAN, 1);
//    //sm_config_set_out_pins(&c, pin_sync_xy, 1);        // pins to use in OUT instr and MOV dest.
//    //sm_config_set_in_pins (&c, pin_sync_xy);           // pins to use in IN instr  and MOV source
//    sm_config_set_jmp_pin(&c, pin_sync_xy);
//
//    sm_config_set_sideset_pins(&c, pin_x);      // pins for 'side setting' of x+ and x-
//    pio_sm_init(pio, sm_x, offset + xy2_100_data_offset_start, &c);
//
//    sm_config_set_sideset_pins(&c, pin_y);      // pins for 'side setting' of y+ and y-
//    pio_sm_init(pio, sm_y, offset + xy2_100_data_offset_start, &c);
//}

void XY2::init_laser ()
{
	uint sm = sm_laser;
	uint32 laser = 1 << pin_laser;
	uint32 syncxy = 1 << pin_sync_xy;

	// laser ON/OFF control pin is controlled by OUT:

    pio_sm_set_pins_with_mask(pio, sm, laser*0, laser);		// initial state: off
	pio_sm_set_pindirs_with_mask(pio, sm, laser*1, laser);	// direction: out
	pio_gpio_init(pio, pin_laser);
	gpio_set_outover(pin_laser,GPIO_OVERRIDE_INVERT); 		// invert output: '1' = ON => pin low


	// xy fifo sync pin is controlled by side-set:

    pio_sm_set_pins_with_mask(pio, sm, syncxy*0, syncxy);		// initial state: 0 (don't care)
	pio_sm_set_pindirs_with_mask(pio, sm, syncxy*1, syncxy);	// direction: out
	pio_gpio_init(pio, pin_sync_xy);

    //pio_sm_set_pins_with_mask(pio, sm_y, 0b1u << pin_sync_xy, 0b1u << pin_sync_xy);
    //pio_sm_set_consecutive_pindirs(pio, sm_y, pin_sync_xy, 1, true/*output*/);
    //pio_gpio_init(pio, pin_sync_xy);


    // load program into pio:
    uint offset = pio_add_program(pio, &xy2_laser_control_program);
    program_base_laser = offset;

    // configure state machines:
    pio_sm_config c = xy2_laser_control_program_get_default_config(offset);
    sm_config_set_clkdiv(&c, float(MAIN_CLOCK) / XY2_SM_CLOCK);
    sm_config_set_out_shift(&c, true/*shift right*/, false /*!autopull*/, 10 /*pull_threshold*/);
	sm_config_set_fifo_join(&c, PIO_FIFO_JOIN_TX);
    sm_config_set_mov_status(&c, STATUS_TX_LESSTHAN, 1);
    sm_config_set_out_pins(&c, pin_laser, 1);        	// pins to use in OUT
    sm_config_set_sideset_pins(&c, pin_sync_xy);      	// pins for 'side setting' of sync_xy
    pio_sm_init(pio, sm_laser, offset + xy2_laser_control_offset_start, &c);
}

//void XY2::init_laser ()
//{
//	gpio_init(PIN_XY2_LASER);
//	gpio_set_dir(PIN_XY2_LASER, GPIO_OUT);
//	gpio_set_outover(PIN_XY2_LASER,GPIO_OVERRIDE_INVERT); // invert output => on = low
//	gpio_put(PIN_XY2_LASER,off);
//}



void XY2::start ()
{
    pio_sm_set_enabled(pio, sm_laser, false);
    pio_sm_set_enabled(pio, sm_clock, false);
    pio_sm_set_enabled(pio, sm_x, false);
    pio_sm_set_enabled(pio, sm_y, false);

    pio_sm_clear_fifos(pio, sm_laser);
    pio_sm_clear_fifos(pio, sm_x);
    pio_sm_clear_fifos(pio, sm_y);

    //set first instr for sm:
    //pio_sm_exec(pio, sm, pio_encode_wait_gpio(trigger_level, trigger_pin));

    // Enable multiple PIO state machines synchronizing their clock dividers
    pio_enable_sm_mask_in_sync(pio, (1<<sm_clock)+(1<<sm_x)+(1<<sm_y)+(1<<sm_laser));
}


void XY2::move_to (const Point& dest)
{
	Point p = Point(x0,y0);
	Dist dist = dest - p;
    FLOAT line_length = dist.length();	// SQRT
    FLOAT speed = SCANNER_MAX_STEP;
    Dist step = dist * (speed / line_length);

    while (line_length > speed)
    {
        line_length -= speed;
    	p = p + step;
        send_data_blocking(p, off);
    }

    if (p != dest)
    {
        send_data_blocking(dest, off);
    }

	for (int delay = LASER_JUMP_DELAY; delay; delay--)
	{
		send_data_blocking(dest, off);
	}

    x0 = dest.x;
    y0 = dest.y;
}

void XY2::draw_to (const Point& dest, FLOAT speed)
{
	Point p = Point(x0,y0);
	Dist dist = dest - p;
    FLOAT line_length = dist.length();		// SQRT
    Dist step = dist * (speed / line_length);

	int laser_on = 0 - LASER_ON_DELAY;

    while (line_length > speed)
    {
    	p = p + step;
        line_length -= speed;
        send_data_blocking(p, ++laser_on > 0);
    }

	int laser_off = 0 - LASER_OFF_DELAY;

	while (++laser_off <= 0)
	{
		send_data_blocking(dest, ++laser_on > 0);
	}

    x0 = dest.x;
    y0 = dest.y;
}

void XY2::drawLine (const Point& start, const Point& dest, FLOAT speed)
{
	move_to(start);
	draw_to(dest,speed);
}

void XY2::drawEllipse (const Rect& bbox, uint steps, uint i, FLOAT speed)
{
	FLOAT fx = bbox.width()/2;
	FLOAT fy = bbox.height()/2;

	drawPolyLine(steps+1,[&fx,&fy,&i,&steps]()
	{
		FLOAT angle = 2*pi * i++ / steps;
		return Point(fx*cos(angle),fy*sin(angle));
	});
}

void XY2::drawPolyLine (uint count, std::function<Point()> nextPoint, FLOAT speed)
{
	assert (count >= 2);

	// get starting point and jump to it:
	Point p = nextPoint();
	move_to(p);

	int laser_on = 0 - LASER_ON_DELAY;

	while (--count > 0)
	{
		Point dest = nextPoint();

		Dist dist = dest - p;
	    FLOAT line_length = dist.length();		// SQRT
	    Dist step = dist * (speed / line_length);

	    while (line_length > speed)
	    {
	    	p = p + step;
	        line_length -= speed;
	        send_data_blocking(p, ++laser_on > 0);
	    }

		if (count != 1)
		{
			// if middle point (not last point)
			// wait polygonDelay:

		    #define mid_delay 0	// option: TODO
			for (int i=0; i<mid_delay; i++)
			{
		        send_data_blocking(p, ++laser_on > 0);
			}
		}
		else
		{
			// behind last point:
			// wait laserOffDelay before switching off the laser

			int laser_off = 0 - LASER_OFF_DELAY;

			while (++laser_off <= 0)
			{
				send_data_blocking(dest, ++laser_on > 0);
			}

		    x0 = dest.x;
		    y0 = dest.y;
		}
	}
}

void XY2::drawRect (Rect bbox, FLOAT speed)
{
	// find nearest starting point:
	Point m = bbox.center();
	if (x0 > m.x) std::swap(bbox.left,bbox.right);
	if (y0 > m.y) std::swap(bbox.top,bbox.bottom);

	// draw it:
	move_to(bbox.top_left());
	draw_to(bbox.top_right());
	draw_to(bbox.bottom_right());
	draw_to(bbox.bottom_left());
	draw_to(bbox.top_left());
}

static inline void rotate_cw (Point& p, FLOAT sin, FLOAT cos)
{
	FLOAT px = p.x * cos + p.y * sin;
	FLOAT py = p.y * cos - p.x * sin;
	p.x = px;
	p.y = py;
}

#ifdef XY2_IMPLEMENT_CHECKER_BOARD_DEMO
void XY2::drawCheckerBoard (const Rect& bbox, uint count, FLOAT speed)
{
    for (uint i=0; i<=count; i++)
    {
        int y = bbox.bottom + bbox.height() * i / count;
        drawLine(Point(i & 1 ? bbox.right : bbox.left,y), Point(i & 1 ? bbox.left : bbox.right,y));
    }
    for (uint i=0; i<=count; i++)
    {
        int x = bbox.left + bbox.width() * i / count;
        drawLine(Point(x, i & 1 ? bbox.top : bbox.bottom), Point(x, i & 1 ? bbox.bottom : bbox.top));
    }
}
#endif

#ifdef XY2_IMPLEMENT_ANALOGUE_CLOCK_DEMO
void XY2::drawClock (const Rect& bbox, uint32 second, FLOAT speed)
{
	assert(bbox.center() == Point(0,0));

	uint minute = second / 60;
	uint hour = minute / 60;
	second %= 60;
	minute %= 60;
	hour   %= 12;

	uint i;
	FLOAT rad;
	Point triangle[3];
	const FLOAT radius = bbox.width()/2;

	const FLOAT
		O_RING_DIA = 1.0,   // outer ring diameter
		I_RING_DIA = 0.05,  // clock axis diameter
		HM_O_DIA  = 0.95,   // hour markers outer diameter
		HM_I_DIA  = 0.80,   // hour markers inner diameter
		SH_LENGTH = 0.90, SH_I_DIA = -0.2,                   // seconds handle length and inner length
		MH_LENGTH = 0.70, MH_I_DIA = -0.15, MH_WIDTH = 0.05, // minute handle
		HH_LENGTH = 0.55, HH_I_DIA = -0.15, HH_WIDTH = 0.05; // hour handle

	// draw outer ring:
	drawEllipse(bbox * O_RING_DIA, 12*2, 3*2);   // 3 = start at 12 o'clock

	// draw full hour markers for 1 to 11 o'clock:
	for (i=1; i<=11; i++)
	{
		rad = pi*2 / 12 * i;
		Point direction { radius * sin(rad), radius * cos(rad) };
		drawLine(direction * HM_O_DIA, direction * HM_I_DIA);
	}

	// draw hour marker for 12 o'clock: (a triangle)
	FLOAT w = radius * (HM_O_DIA - HM_I_DIA) / 3;   // width/2 of the triangle
	triangle[0] = Point( 0, radius * HM_I_DIA);      // tip
	triangle[1] = Point(-w, radius * HM_O_DIA);
	triangle[2] = Point(+w, radius * HM_O_DIA);

	i=0; drawPolyLine(4, [&i,&triangle](){ return triangle[i++ % 3]; }, speed/2); // start drawing at tip

	// draw the handles starting and ending at the clock center to minimize jumps:

	// draw second handle:
	rad = pi*2 / 60 * second;
	Point direction { radius * sin(rad), radius * cos(rad) };
	drawLine(direction * SH_I_DIA,  direction * SH_LENGTH);
	drawLine(direction * SH_LENGTH, direction * SH_I_DIA);

	// draw minute handle:
	triangle[0] = Point( 0.0,      MH_LENGTH );		// tip
	triangle[1] = Point(-MH_WIDTH, MH_I_DIA);
	triangle[2] = Point(+MH_WIDTH, MH_I_DIA);

	rad = pi*2 / 60 * (minute + second/FLOAT(60.0));
	FLOAT sinus = sin(rad) * radius;
	FLOAT cosin = cos(rad) * radius;
	for (i=0; i<3; i++) { rotate_cw(triangle[i],sinus,cosin); }
	i=0; drawPolyLine(4,[&i,&triangle](){ return triangle[++i % 3]; }); // start drawing at base

	// draw hour handle:
	triangle[0] = Point( 0.0,      HH_LENGTH );		// tip
	triangle[1] = Point(-HH_WIDTH, HH_I_DIA);
	triangle[2] = Point(+HH_WIDTH, HH_I_DIA);

	rad = pi*2 / 12 * (hour + minute/FLOAT(60.0));
	sinus = sin(rad) * radius;
	cosin = cos(rad) * radius;
	for (i=0; i<3; i++) { rotate_cw(triangle[i],sinus,cosin); }
	i=0; drawPolyLine(4,[&i,&triangle](){ return triangle[++i % 3]; }); // start drawing at base

	// draw inner ring: (axis)
	drawEllipse(bbox * I_RING_DIA, 8, speed/2);
}
#endif

#ifdef XY2_IMPLEMENT_LISSAJOUS_DEMO
void XY2::drawLissajous (
		FLOAT width, FLOAT height, FLOAT min_f_ratio, FLOAT max_f_ratio,
		uint32 steps_per_rot, uint32 rots_per_sweep, FLOAT speed)
{
	// draw infinite line:
	// don't bother with start and end

	steps_per_rot = minmax(4u, steps_per_rot, 1000u);				// steps for full circle in x direction
	rots_per_sweep = minmax(4u, rots_per_sweep, 1000000u);			// circles for full rotation min_fy/fx <-> max_fy/fx
	min_f_ratio = minmax(FLOAT(0.5), min_f_ratio, FLOAT(10.05));	// min. fy/fx ratio
	max_f_ratio = minmax(min_f_ratio, max_f_ratio, FLOAT(10.05));	// max. fy/fx ratio

	struct PolyData {
		FLOAT x_scale, y_scale, x_step, x_angle, y_angle;
		FLOAT f_ratio_center, f_ratio_angle_step, f_ratio_angle, f_ratio_scale;
	} p {
		.x_scale = width / 2,
		.y_scale = height / 2,
		.x_step  = 2*pi / steps_per_rot,
		.x_angle = 0.0,
		.y_angle = 0.0,
		.f_ratio_center = sqrt(min_f_ratio * max_f_ratio),
		.f_ratio_angle_step = 2*pi / steps_per_rot / rots_per_sweep,
		.f_ratio_angle = 0.0,
		.f_ratio_scale = (max_f_ratio-min_f_ratio)/2,
	};

	drawPolyLine(1000*100000, [&p]()
	{
		p.f_ratio_angle += p.f_ratio_angle_step;
		if (p.f_ratio_angle >= pi) p.f_ratio_angle -= pi*2;
		FLOAT f_ratio = p.f_ratio_center + sinf(p.f_ratio_angle) * p.f_ratio_scale;

		p.x_angle += p.x_step;
		if (p.x_angle >= pi) p.x_angle -= pi*2;
		p.y_angle += p.x_step * f_ratio;
		if (p.y_angle >= pi) p.y_angle -= pi*2;

		FLOAT x = sinf(p.x_angle) * p.x_scale;
		FLOAT y = cosf(p.y_angle) * p.y_scale;

		return Point{x,y};
	});
}
#endif




