#include <stdio.h>
#include <math.h>
#include "pico/stdlib.h"
#include "pico/binary_info.h"
#include "pico/multicore.h"         // für critical_section für DI / EI
#include "hardware/gpio.h"
#include "hardware/dma.h"
#include "hardware/pio.h"
#include "hardware/pio_instructions.h"
#include "hardware/structs/pwm.h"   // for example waveforms to analyse
#include "XY2-100.pio.h"
#include "cdefs.h"
#include "LogicAnalyzer.h"
#include "XY2.h"
#include "utilities.h"
extern "C" {
#include "hardware/rtc.h"
}
//#include "pico/util/datetime.h"


namespace kio {


const uint CAPTURE_PIN_BASE     = PIN_XY2_CLOCK;
const uint CAPTURE_PIN_COUNT    = 8;
const uint CAPTURE_TRIGGER_PIN  = PIN_XY2_CLOCK;
const uint CAPTURE_N_SAMPLES    = 40 * 25;


void start_demo_pwm(int pin1, int num_pins)
{
    for (int pin = pin1; pin < pin1+num_pins; pin++)
    {
        gpio_set_function(pin, GPIO_FUNC_PWM);
    }

    for (int pin = pin1 & 0xfe; pin < pin1+num_pins; pin+=2)
    {
        int slice_no = (pin>>1) & 7;
        pwm_slice_hw_t* slice = &pwm_hw->slice[slice_no];

        //printf ("configuring slice %i\n",slice_no);

        switch(slice_no)
        {
        case 0:
        case 4:
            // Topmost value of 3: count from 0 to 3 and then wrap, so period is 4 cycles
            slice->top = 3;
            // Divide frequency by two to slow things down a little
            slice->div = 4 << PWM_CH0_DIV_INT_LSB;
            // Set channel A to be high for 1 cycle each period (duty cycle 1/4) and
            // channel B for 3 cycles (duty cycle 3/4)
            slice->cc = (1 << PWM_CH0_CC_A_LSB) | (3 << PWM_CH0_CC_B_LSB);
            // Enable this PWM slice
            //slice->csr = PWM_CH0_CSR_EN_BITS;
            break;
        case 1:
        case 5:
            slice->top = 4;
            slice->div = 2 << PWM_CH1_DIV_INT_LSB;
            slice->cc = (1 << PWM_CH1_CC_A_LSB) | (3 << PWM_CH1_CC_B_LSB);
            //slice->csr = PWM_CH1_CSR_EN_BITS;
            break;
        default:    // constants are the same for all PWMs
            slice->top = 5;
            slice->div = 1 << PWM_CH2_DIV_INT_LSB;
            slice->cc = (1 << PWM_CH2_CC_A_LSB) | (3 << PWM_CH2_CC_B_LSB);
            //slice->csr = PWM_CH2_CSR_EN_BITS;
            break;
        }
    }

    critical_section_t cs;
    critical_section_init(&cs);
    critical_section_enter_blocking(&cs);

    for (int i = pin1>>1; i < (pin1+num_pins+1)>>1; i++)
    {
        int slice_no = i&7;
        pwm_slice_hw_t* slice = &pwm_hw->slice[slice_no];
        slice->csr = PWM_CH0_CSR_EN_BITS;   // constants are the same for all PWMs
    }

    critical_section_exit(&cs);
}



int main()
{
    //bi_decl(bi_program_description("This is a test binary."));
    //bi_decl(bi_1pin_with_name(LED_PIN, "On-board LED"));

    stdio_init_all();
    gpio_init(LED_PIN); gpio_set_dir(LED_PIN, GPIO_OUT);
    gpio_init(LED_BEAT); gpio_set_dir(LED_BEAT, GPIO_OUT); gpio_put(LED_BEAT,0);
    gpio_init(LED_IDLE); gpio_set_dir(LED_IDLE, GPIO_OUT); gpio_put(LED_IDLE,0);
    gpio_init(LED_BUSY); gpio_set_dir(LED_BUSY, GPIO_OUT); gpio_put(LED_BUSY,1);
    gpio_init(LED_ERROR); gpio_set_dir(LED_ERROR, GPIO_OUT); gpio_put(LED_ERROR,0);


    printf("XY2-100 LaserScanner Example\n");
    printf("SCANNER_MAX_STEP = %u\n", SCANNER_MAX_STEP);
    //while(getchar()!=13);

    char charbuffer[256] = "";
    printf("please enter size (%%), HH:MM:SS, YYYY/MM/DD (all optional)\n> ");
    uint i=0;
    while(getchar_timeout_us(0)>=0){}
    for(;;)
    {
        char c = getchar_timeout_us(60*1000*1000);
        if (c<0 || c==13) break;
        if ((c==8 || c==127) && i>0) { printf("\x08 \x08"); i--; continue; }
        if (i<NELEM(charbuffer)-1) { putchar(c); charbuffer[i++] = c; }
    }
    charbuffer[i] = 0;
    putchar('\n');

    i = 0;
    int size    = parseInteger(charbuffer, i, 5, 30, 100);
    uint hour   = parseInteger(charbuffer, i, 0, 10, 23);
    uint minute = parseInteger(charbuffer, i, 0,  0, 59);
    uint second = parseInteger(charbuffer, i, 0,  0, 59);
    uint year   = parseInteger(charbuffer, i, 0, 2020, 2999);
    uint month  = parseInteger(charbuffer, i, 1,  7, 12);
    uint day    = parseInteger(charbuffer, i, 1, 20, 31);


    // Start on Friday 5th of June 2020 15:45:00
    datetime_t t = {
            .year  = int16(year),
            .month = int8(month),
            .day   = int8(day),
            .dotw  = 0, // 0 is Sunday. TODO...
            .hour  = int8(hour),
            .min   = int8(minute),
            .sec   = int8(second)
    };

    // Start the RTC
    rtc_init();
    rtc_set_datetime(&t);


    //int size  = parseInteger(m_conBuffer, idx, 5, 30, 100);
    FLOAT fmin = 1.49;//parseFloat  (charbuffer, i, 0.95, 1.49, 10.05);
    FLOAT fmax = 1.51;//parseFloat  (charbuffer, i, fmin, 1.51, 10.05);
    int steps  = 100; //parseInteger(charbuffer, i, 10, 100, 1000);
    int rots   = 5000;//parseInteger(charbuffer, i, 100, 5000, 1000000);
    printf("Lissajous demo:\n");
    printf("size=%i%%, fmin=%.3f, fmax=%.3f, steps/rot=%i, rots/sweep=%i\n", size,double(fmin),double(fmax),steps,rots );


    // for logic analyzer:
    PIO  LA_pio = pio1;
    uint LA_sm = 0;
    uint LA_dma_chan = 0;
    uint32_t capture_buf[(CAPTURE_PIN_COUNT * CAPTURE_N_SAMPLES + 31) / 32];

    printf("Starting logic analyzer\n");
    logic_analyser_init(LA_pio, LA_sm, CAPTURE_PIN_BASE, CAPTURE_PIN_COUNT, MAIN_CLOCK/8000000.0);
    logic_analyser_arm(LA_pio, LA_sm, LA_dma_chan, capture_buf, (CAPTURE_PIN_COUNT * CAPTURE_N_SAMPLES + 31) / 32, CAPTURE_TRIGGER_PIN, true);


    printf("Starting XY2-100 interface\n");
    XY2 xy2;
    xy2.start();

    //x0 = y0 = 0;
    //xy2_send_data_blocking(XY_pio,sm_x,sm_y,x0,y0);
    //xy2.send_data_blocking(x0,y0);

    int w = SCANNER_WIDTH*size/100;
    int h = w;
    Rect bbox{FLOAT(h/2), FLOAT(-w/2), FLOAT(-h/2), FLOAT(w/2)};

    for (int i=0;i<10;i++)
    {
        xy2.drawRect(bbox);
        //xy2.draw_rect (-w/2, -h/2, w, h);
    }

    dma_channel_wait_for_finish_blocking(LA_dma_chan);
    print_capture_buf(capture_buf, CAPTURE_PIN_BASE, CAPTURE_PIN_COUNT, CAPTURE_N_SAMPLES);


    bool led_on=0;
    xy2.drawCheckerBoard(bbox, 4);

    while (1)
    {
        // puts("Hello World\n");
        //sleep_ms(125);
        gpio_put(LED_PIN, led_on=!led_on);

        rtc_get_datetime(&t);
        uint32 second = t.hour*60*60 + t.min*60 + t.sec;
        //xy2.drawClock (bbox, second);

        xy2.drawLissajous (w,h, fmin, fmax, steps, rots);
    }
}

} // namespace

int main()
{
    return kio::main();
}








